#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018, Blue Brain Project
#                     Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#
# This file is part of Rockets <https://github.com/BlueBrain/Rockets>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.
"""Asynchronous client implementation for asyncio event loop processing of JSON-RPC messages."""
import asyncio
import json
import sys
from functools import reduce

import websockets
from jsonrpc.jsonrpc2 import JSONRPC20Request
from rx import Observable

from .notification import Notification
from .request import Request
from .request_error import INVALID_REQUEST
from .request_error import RequestError
from .request_error import SOCKET_CLOSED_ERROR
from .request_progress import RequestProgress
from .request_task import RequestTask
from .response import Response
from .utils import is_json_rpc_notification
from .utils import is_json_rpc_response
from .utils import is_progress_notification
from .utils import set_ws_protocol


class AsyncClient:
    """Asynchronous client implementation for asyncio event loop processing of JSON-RPC messages."""

    def __init__(self, url, subprotocols=None, loop=None):
        """
        Initialize the state of the client.

        Convert the URL to a proper format. Does not establish the websocket connection yet. This
        will be postponed to the first notify or request.

        :param str url: The address of the Rockets server.
        :param list subprotocols: The websocket protocols to use
        :param asyncio.AbstractEventLoop loop: Event loop where this client should run in
        """
        self.url = set_ws_protocol(url)
        """The address of the connected Rockets server."""

        if not subprotocols:
            subprotocols = ["rockets"]
        self._subprotocols = subprotocols

        self._ws = None

        self.loop = loop
        """The event loop where this client is running in."""
        if not self.loop:
            self.loop = asyncio.get_event_loop()

        def _ws_loop(observer):
            """Internal: synchronous wrapper for async _ws_loop"""
            asyncio.ensure_future(self._ws_loop(observer), loop=self.loop)

        # pylint: disable=E1101
        self.ws_observable = Observable.create(_ws_loop).publish().auto_connect()
        """The websocket stream as an rx observable to subscribe to it."""
        # pylint: enable=E1101

        def _json_filter(value):
            try:
                json.loads(value)
                return True
            except ValueError:  # pragma: no cover
                return False

        # filter everything that is not JSON
        self._json_stream = self.ws_observable.filter(_json_filter).map(json.loads)

        def _notifications_filter(value):
            return is_json_rpc_notification(value) and not is_progress_notification(
                value
            )

        self.notifications = self._json_stream.filter(_notifications_filter).map(
            lambda x: Notification.from_json(json.dumps(x))
        )
        """The rx observable to subscribe to notifications from the server."""

    def connected(self):
        """
        Returns the connection state of this client.

        :return: true if the websocket is connected to the Rockets server.
        :rtype: bool
        """
        return bool(self._ws and self._ws.open)

    async def connect(self):
        """Connect this client to the Rockets server"""
        if self.connected():
            return

        self._ws = await websockets.connect(
            self.url, subprotocols=self._subprotocols, loop=self.loop
        )

    async def disconnect(self):
        """Disconnect this client from the Rockets server."""
        if not self.connected():
            return

        await self._ws.close()

    async def send(self, message):
        """
        Send any message to the connected Rockets server.

        :param str message: The message to send
        """
        await self.connect()
        await self._ws.send(message)

    async def notify(self, method, params):
        """
        Invoke an RPC on the Rockets server without expecting a response.

        :param str method: name of the method to invoke
        :param str params: params for the method
        """
        notification = Notification(method, params)
        await self.send(notification.json)

    async def request(self, method, params=None):
        """
        Invoke an RPC on the Rockets server and returns its response.

        :param str method: name of the method to invoke
        :param dict params: params for the method
        :return: future object
        :rtype: :class:`asyncio.Future`
        """
        if params and not isinstance(params, (list, tuple, dict)):
            params = [params]
        request = Request(method, params)
        try:
            response_future = self.loop.create_future()

            request_id = request.request_id()
            await self.connect()
            self._setup_response_filter(response_future, request_id)
            self._setup_progress_filter(response_future, request_id)

            await self.send(request.json)
            await response_future
            return response_future.result()
        except asyncio.CancelledError:
            await self.notify("cancel", {"id": request_id})

    async def batch(self, requests):
        """
        Invoke a batch RPC on the Rockets server and return its response(s).

        :param list requests: list of requests and/or notifications to send as batch
        :return: future object with list of responses
        :rtype: :class:`asyncio.Future`
        :raises RequestError: if methods and/or params are not a list
        :raises RequestError: if methods are empty
        """
        if not requests:
            raise INVALID_REQUEST

        # is valid for both, requests and notifications
        for request in requests:
            if not isinstance(request, JSONRPC20Request):
                raise INVALID_REQUEST

        request_ids = list()
        for request in requests:
            if isinstance(request, Request):
                request_ids.append(request.request_id())

        try:

            def _data(x):
                return x.data

            request = Request.from_data(list(map(_data, requests)))

            response_future = self.loop.create_future()

            await self.connect()
            self._setup_batch_response_filter(response_future, request_ids)
            self._setup_batch_progress_filter(response_future, request_ids)

            await self.send(request.json)
            await response_future
            return response_future.result()
        except asyncio.CancelledError:
            for request_id in request_ids:
                await self.notify("cancel", {"id": request_id})

    def async_request(self, method, params=None):
        """
        Invoke an RPC on the Rockets server and return the :class:`RequestTask`.

        :param str method: name of the method to invoke
        :param dict params: params for the method
        :return: :class:`RequestTask` object
        :rtype: :class:`RequestTask`
        """
        self.loop.set_task_factory(lambda loop, coro: RequestTask(coro=coro, loop=loop))

        task = self.request(method, params)
        return asyncio.ensure_future(task, loop=self.loop)

    def async_batch(self, requests):
        """
        Invoke a batch RPC on the Rockets server and return the :class:`RequestTask`.

        :param list requests: list of requests and/or notifications to send as batch
        :return: :class:`RequestTask` object
        :rtype: :class:`RequestTask`
        """
        self.loop.set_task_factory(lambda loop, coro: RequestTask(coro=coro, loop=loop))

        task = self.batch(requests)
        return asyncio.ensure_future(task, loop=self.loop)

    async def _ws_loop(self, observer):
        """Internal: The loop for feeding an rxpy observer."""
        try:
            await self.connect()
            if sys.version_info < (3, 6):
                while True:  # pragma: no cover
                    message = await self._ws.recv()
                    observer.on_next(message)
            else:  # pragma: no cover
                async for message in self._ws:
                    observer.on_next(message)
                observer.on_completed()
        except websockets.ConnectionClosed:  # pragma: no cover
            observer.on_completed()

    def _setup_response_filter(self, response_future, request_id):
        def _response_filter(value):
            return is_json_rpc_response(value) and value["id"] == request_id

        def _to_response(value):
            response = Response.from_json(json.dumps(value))
            if response.result:
                return response.result
            return response.error

        def _on_next(value):
            if not response_future.done():
                if isinstance(value, dict) and "code" in value:
                    response_future.set_exception(RequestError(**value))
                else:
                    response_future.set_result(value)

        def _on_completed():

            if not response_future.done():
                response_future.set_exception(SOCKET_CLOSED_ERROR)

        self._json_stream.filter(_response_filter).take(1).map(_to_response).subscribe(
            on_next=_on_next, on_completed=_on_completed
        )

    def _setup_batch_response_filter(self, response_future, request_ids):
        def _response_filter(value):
            if not isinstance(value, list):
                return False
            for response in value:
                if (
                    not is_json_rpc_response(response)
                    or response["id"] not in request_ids
                ):
                    return False  # pragma: no cover
            return True

        def _to_response(value):
            responses = [Response.from_json(json.dumps(i)) for i in value]
            return responses

        def _on_next(value):
            if not response_future.done():
                response_future.set_result(value)

        def _on_completed():
            if not response_future.done():
                response_future.set_exception(SOCKET_CLOSED_ERROR)

        self._json_stream.filter(_response_filter).take(1).map(_to_response).subscribe(
            on_next=_on_next, on_completed=_on_completed
        )

    def _setup_progress_filter(self, response_future, request_id):
        task = asyncio.Task.current_task()
        if task and isinstance(task, RequestTask):

            def _progress_filter(value):
                return (
                    is_progress_notification(value)
                    and value["params"]["id"] == request_id
                )

            def _to_progress(value):
                progress = Request.from_data(value).params
                return RequestProgress(progress["operation"], progress["amount"])

            progress_observable = (
                self._json_stream.filter(_progress_filter)
                .map(_to_progress)
                .subscribe(task._call_progress_callbacks)  # pylint: disable=W0212
            )

            def _done_callback(future):  # pylint: disable=W0613
                progress_observable.dispose()

            response_future.add_done_callback(_done_callback)

    def _setup_batch_progress_filter(self, response_future, request_ids):
        task = asyncio.Task.current_task()
        items = dict()
        if task and isinstance(task, RequestTask):

            def _progress_filter(value):
                return (
                    is_progress_notification(value)
                    and value["params"]["id"] in request_ids
                )

            def _to_progress(value):
                progress = Request.from_data(value).params
                items[progress["id"]] = progress["amount"]
                total = reduce((lambda x, value: x + value), items.values())
                return RequestProgress("Batch request", total / len(request_ids))

            progress_observable = (
                self._json_stream.filter(_progress_filter)
                .map(_to_progress)
                .subscribe(task._call_progress_callbacks)  # pylint: disable=W0212
            )

            def _done_callback(future):  # pylint: disable=W0613
                progress_observable.dispose()

            response_future.add_done_callback(_done_callback)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018, Blue Brain Project
#                     Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#
# This file is part of Rockets <https://github.com/BlueBrain/Rockets>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.
"""A JSON-RPC 2.0 request"""
from jsonrpc.jsonrpc2 import JSONRPC20Request

from .utils import random_string


class Request(JSONRPC20Request):
    """A JSON-RPC 2.0 request"""

    _id_generator = random_string()

    def __init__(self, method, params=None):
        super().__init__(
            method=method,
            params=params,
            _id=next(Request._id_generator),
            is_notification=False,
        )

    def request_id(self):
        """Return the request ID"""
        return super()._id



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018, Blue Brain Project
#                     Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#
# This file is part of Rockets <https://github.com/BlueBrain/Rockets>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.
"""Client that support synchronous and asynchronous usage of the :class:`AsyncClient`."""
import asyncio
from threading import Thread

from .async_client import AsyncClient
from .utils import copydoc


class Client:
    """Client that support synchronous usage of the :class:`AsyncClient`."""

    def __init__(self, url, subprotocols=None, loop=None):
        """
        Setup the :class:`AsyncClient` for synchronous usage.

        In case the given loop is running, uses a threaded client to achieve synchronous execution.

        :param str url: The address of the Rockets server.
        :param list subprotocols: The websocket protocols to use
        :param asyncio.AbstractEventLoop loop: Event loop where this client should run in
        """
        if not loop:
            loop = asyncio.get_event_loop()

        if loop.is_running():
            thread_loop = asyncio.new_event_loop()

            def _start_background_loop(loop):
                asyncio.set_event_loop(loop)
                loop.run_forever()

            self._thread = Thread(target=_start_background_loop, args=(thread_loop,))
            self._thread.daemon = True
            self._thread.start()

            self._client = AsyncClient(url, subprotocols=subprotocols, loop=thread_loop)
        else:
            self._thread = None
            self._client = AsyncClient(url, subprotocols=subprotocols, loop=loop)

        self.url = self._client.url
        """The address of the connected Rockets server."""

        self.ws_observable = self._client.ws_observable
        """The websocket stream as an rx observable to subscribe to it."""

        self.notifications = self._client.notifications
        """The rx observable to subscribe to notifications from the server."""

    @copydoc(AsyncClient.connected)
    def connected(self):  # noqa: D102 pylint: disable=missing-docstring
        return self._client.connected()

    @copydoc(AsyncClient.connect)
    def connect(self):  # noqa: D102 pylint: disable=missing-docstring
        self._call_sync(self._client.connect())

    @copydoc(AsyncClient.disconnect)
    def disconnect(self):  # noqa: D102 pylint: disable=missing-docstring
        self._call_sync(self._client.disconnect())

    @copydoc(AsyncClient.send)
    def send(self, message):  # noqa: D102 pylint: disable=missing-docstring
        self._call_sync(self._client.send(message))

    @copydoc(AsyncClient.notify)
    def notify(
        self, method, params=None
    ):  # noqa: D102 pylint: disable=missing-docstring
        self._call_sync(self._client.notify(method, params))

    @copydoc(AsyncClient.request)
    def request(
        self, method, params=None, response_timeout=None
    ):  # noqa: D102,D205 pylint: disable=C0111,W9011,W9012,W9015,W9016
        """
        :param int response_timeout: number of seconds to wait for the response
        :raises TimeoutError: if request was not answered within given response_timeout
        """
        return self._call_sync(self._client.request(method, params), response_timeout)

    @copydoc(AsyncClient.batch)
    def batch(
        self, requests, response_timeout=None
    ):  # noqa: D102,D205 pylint: disable=C0111,W9011,W9012,W9015,W9016
        """
        :param int response_timeout: number of seconds to wait for the response
        :raises TimeoutError: if request was not answered within given response_timeout
        """
        return self._call_sync(self._client.batch(requests), response_timeout)

    def _call_sync(self, original_function, response_timeout=None):
        if not self._thread and self._client.loop.is_running():
            raise RuntimeError("Unknown working environment")
        if self._thread:
            future = asyncio.run_coroutine_threadsafe(
                original_function, self._client.loop
            )
            return future.result(response_timeout)
        return self._client.loop.run_until_complete(original_function)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018, Blue Brain Project
#                     Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#
# This file is part of Rockets <https://github.com/BlueBrain/Rockets>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.
"""The version of the rockets package"""
from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution

try:
    VERSION = get_distribution("rockets").version
except DistributionNotFound:  # pragma: no cover
    VERSION = "rockets-local"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018, Blue Brain Project
#                     Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#
# This file is part of Rockets <https://github.com/BlueBrain/Rockets>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""Reports the operation and amount of a request."""


class RequestProgress:
    """Reports the operation and amount of a request."""

    def __init__(self, operation, amount):
        self.operation = operation
        self.amount = amount

    def __str__(self):
        """
        Print progress as string

        :return: tuple(operation,amount)
        :rtype: str
        """
        return str((self.operation, self.amount))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018, Blue Brain Project
#                     Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#
# This file is part of Rockets <https://github.com/BlueBrain/Rockets>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""Extend asyncio.Task to add callbacks for progress reporting while the request is not done."""

import asyncio


class RequestTask(asyncio.Task):
    """Extend asyncio.Task to add callbacks for progress reporting while the request is not done."""

    def __init__(self, coro, *, loop=None):
        super().__init__(coro=coro, loop=loop)
        self._progress_callbacks = []

    def add_progress_callback(self, fn):
        """
        Add a callback to be run everytime a progress update arrives.

        The callback is called with a single argument - the :class:`RequestProgress` object.
        """
        self._progress_callbacks.append(fn)

    def _call_progress_callbacks(self, value):
        """Internal: Calls registered progress callbacks."""
        for callback in self._progress_callbacks:
            callback(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018, Blue Brain Project
#                     Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#
# This file is part of Rockets <https://github.com/BlueBrain/Rockets>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""setup.py"""
import os
from setuptools import setup

try: # for pip >= 10
    from pip._internal.req import parse_requirements
    from pip._internal.download import PipSession
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements
    from pip.download import PipSession
from optparse import Option

BASEDIR = os.path.dirname(os.path.abspath(__file__))


def parse_reqs(reqs_file):
    ''' parse the requirements '''
    options = Option("--workaround")
    options.skip_requirements_regex = None
    options.isolated_mode = True
    install_reqs = parse_requirements(reqs_file, options=options, session=PipSession())
    return [str(ir.req) for ir in install_reqs]


REQS = parse_reqs(os.path.join(BASEDIR, "requirements.txt"))

# read the contents of README.md
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md')) as f:
    long_description = f.read()

setup(
    packages=['rockets'],
    install_requires=REQS,
    long_description=long_description,
    long_description_content_type='text/markdown'
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018, Blue Brain Project
#                     Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#
# This file is part of Rockets <https://github.com/BlueBrain/Rockets>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.
"""A small client for Rockets using JSON-RPC as communication contract over a WebSocket."""
from .async_client import AsyncClient
from .client import Client
from .notification import Notification
from .request import Request
from .request_error import RequestError
from .request_progress import RequestProgress
from .request_task import RequestTask
from .response import Response
from .version import VERSION as __version__

__all__ = [
    "AsyncClient",
    "Client",
    "Notification",
    "Request",
    "RequestError",
    "RequestProgress",
    "RequestTask",
    "Response",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018, Blue Brain Project
#                     Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#
# This file is part of Rockets <https://github.com/BlueBrain/Rockets>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.
"""Utils for the client"""
from random import choice
from string import ascii_lowercase
from string import digits

HTTP = "http://"
HTTPS = "https://"
WS = "ws://"
WSS = "wss://"


def set_ws_protocol(url):
    """
    Set the WebSocket protocol according to the resource url.

    :param str url: Url to be checked
    :return: Url preprend with ws for http, wss for https for ws if no protocol was found
    :rtype: str
    """
    if url.startswith(WS) or url.startswith(WSS):
        return url
    if url.startswith(HTTP):
        return url.replace(HTTP, WS, 1)
    if url.startswith(HTTPS):
        return url.replace(HTTPS, WSS, 1)
    return WS + url


def copydoc(fromfunc, sep="\n"):
    """
    Copy the docstring of `fromfunc`

    :return: Decorator to use on any func to copy the docstring from `fromfunc`
    :rtype: obj
    """

    def _decorator(func):
        sourcedoc = fromfunc.__doc__
        if func.__doc__:  # pragma: no cover
            func.__doc__ = sep.join([sourcedoc, func.__doc__])
        else:
            func.__doc__ = sourcedoc
        return func

    return _decorator


def random_string(length=8, chars=digits + ascii_lowercase):
    """
    A random string.

    Not unique, but has around 1 in a million chance of collision (with the default 8
    character length). e.g. 'fubui5e6'

    :param int length: Length of the random string.
    :param str chars: The characters to randomly choose from.
    :return: random string
    :rtype: str
    """
    while True:
        yield "".join([choice(chars) for _ in range(length)])


def is_json_rpc_response(value):
    """Check if the given value is valid JSON-RPC response."""
    return isinstance(value, dict) and "id" in value


def is_json_rpc_notification(value):
    """Check if the given value is valid JSON-RPC notification."""
    return isinstance(value, dict) and "method" in value and "id" not in value


def is_progress_notification(value):
    """Check if the given value is valid Rockets progress notification."""
    return (
        isinstance(value, dict)
        and "method" in value
        and value["method"] == "progress"
        and "params" in value
        and "id" in value["params"]
    )
